/*****************************************************************
this package contains the arb-tree implementation of paper
"title goes here"
for questions, comments and bug reports please contact yufei tao at 
taoyf@cs.ust.hk. in case i change my affiliation please search
"yufei tao" on google to reach my current homepage
may 2003
*****************************************************************/

#include <math.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include "./rtree/rtree.h"
#include "./rtree/rtnode.h"
#include "./rtree/entry.h"
#include "./blockfile/blk_file.h"
#include "./linlist/linlist.h"

/*****************************************************************
this function creates the r-tree in the arb-tree
para:
dsfname: the spatial dataset file name
tfname: the tree file name
dsize: the disk page size
hashnum: # of hash functions
bitlen: # of bits in each function's signature
pcsa: toggles if pcsa is used in the signatures
coded by yufei tao, may 2003
******************************************************************/

void build_rtree(char *_dsfname, char *_tfname, int _dsize, int _hashnum, int _bitlen, bool _pcsa)
{
	printf("really build rtree? NOT B-TREES! (y to cont): ");
	char ans;
	ans=getchar();
	getchar();
	if (ans!='y') return;

	Cache *c=new Cache(0, _dsize);
	RTree *rt=new RTree(_dsfname, _tfname, _dsize, c, 2, _hashnum, _bitlen, _pcsa);
	delete rt;
	delete c;
}

/*****************************************************************
this function creates the b-trees in the arb-tree by inserting the bitmaps
para:
dfname: the spatial dataset file name
sfname: the signature file
tfname: the tree file name
dsize: disk page size
rnum: the number of regions in dfname
coded by yufei tao, may 2003
******************************************************************/

void insert_bitmaps(char *_dfname, char *_sfname, char *_tfname, int _dsize, int _rnum)
{
	printf("really insert bitmaps? (y to cont): ");
	char ans;
	ans=getchar();
	getchar();
	if (ans!='y') return;

	Cache *c=new Cache(0, _dsize);
	RTree *rt=new RTree(_tfname, c);
	rt->create_btrees(_dfname, _sfname, _rnum);
	delete rt;
	delete c;
}

/*****************************************************************
this function creates the b-trees in the arb-tree by inserting the bitmaps
para:
dfname: the spatial dataset file name
sfname: the signature file
tfname: the tree file name
dsize: disk page size
rnum: the number of regions in dfname
coded by yufei tao, may 2003
******************************************************************/

void integrity_check(char *_dfname, char *_sfname, char *_tfname, int _dsize, int _rnum)
{
	FILE *dfp=fopen(_dfname, "r");
	if (!dfp)
		error("could not open the spatial file...\n", true);
	FILE *sfp=fopen(_sfname, "rb");
	if (!sfp)
		error("could not open the signature file...\n", true);

	printf("reading the spatial dataset...\n");
	float *reg=new float[4*_rnum];
	int cnt=0;
	while (!feof(dfp))
	{
		if (cnt==_rnum)
			error("\tthe spatial dataset's cardinality is larger than you thought...\n", true);
		int id; 
		fscanf(dfp, "%d ", &id);
		for (int i=0; i<4; i++)
			fscanf(dfp, "%f ", &reg[4*cnt+i]);
		fscanf(dfp, "\n");
		cnt++;
	}
	printf("\tread %d regions\n", cnt);
	if (cnt!=_rnum)
		printf("\twarning: the spatial dataset's cardinality is smaller than you thought...\n");

	Cache *c=new Cache(0, _dsize);
	RTree *rt=new RTree(_tfname, c);

	unsigned char *bitchars=new unsigned char[rt->bitcharlen];
	unsigned char *thisbitchars=new unsigned char[rt->bitcharlen];
	int last_tm=-1; int tm, rid;
	printf("performing integrity check\n");
	binary_int_read(sfp, &tm);
	while(tm!=-1)
	{
		binary_int_read(sfp, &rid);
		for (int i=0; i<rt->bitcharlen; i++)
			bitchars[i]=fgetc(sfp);

		if (tm>last_tm)
		{
			last_tm=tm;
			printf("\ttimestamp %d...\n", tm);
		}

		rt->get_reg_bitchars(rid, &reg[rid*4], tm, thisbitchars);
		bool same=true;
		for (i=0; i<rt->bitcharlen; i++)
			if (thisbitchars[i]!=bitchars[i])
			{
				same=false; break;
			}
		if (!same)
		{
			printf("error for rid=%d, tm=%d\n", rid, tm);
			exit(1);
		}

		binary_int_read(sfp, &tm);
	}
	printf("passed\n");
	delete rt;
	delete c;

	delete []thisbitchars;
	delete []bitchars;
	delete []reg;
}

/*****************************************************************
this function creates the b-trees in the arb-tree by inserting the bitmaps
para:
tfname: the tree file name
qfname: the query file name
blen: the block lenght
ed: the last timestamp in history
mode: =1 a plain algorithm, =2 a better algorithm
coded by yufei tao, may 2003
******************************************************************/

void aggq_wrkld(char *_tfname, char *_qfname, int _blen, int ed, int _mode)
{
	FILE *qfp=fopen(_qfname, "r");
	if (!qfp)
		error("could not open the query file...\n", true);
	Cache *c=new Cache(0, _blen);
	RTree *rt=new RTree(_tfname, c); 
	
	int cnt=0; 
	float max_re=-1, min_re=1;
	float total_re=0; 
	float total_cost=0;

	while(!feof(qfp) && cnt<=20)
	{
		float qmbr[4];
		for (int i=0; i<4; i++)
			fscanf(qfp, "%f ", &qmbr[i]);
		float qt[2];
		for (i=0; i<2; i++)
			fscanf(qfp, "%f ", &qt[i]);
		int res;
		fscanf(qfp, "%d\n", &res);

		if (res>0)
		{
			int old_cost=c->page_faults;
			float est;
//			if (cnt==1)
			{
				if (_mode==1)
					est=rt->agg_qry(qmbr, qt, ed);
				else
					est=rt->agg_qry2(qmbr, qt, ed);
			}
			int new_cost=c->page_faults;
			total_cost+=new_cost-old_cost;
			float rel_err=fabs(est-res)/(float)res;

			total_re+=rel_err;
			if (rel_err<min_re)
				min_re=rel_err;
			if (rel_err>max_re)
				max_re=rel_err;

			cnt++;
			printf("act=%d, est=%.1f, rel_err=%.3f, cost=%d\n", res, est, rel_err*100, new_cost-old_cost);
		}
	}
	total_re/=cnt;
	printf("min re=%.3f%, max re=%.3f, avg re=%.3f\n", min_re*100, max_re*100, total_re*100);
	printf("average cost=%.3f\n", total_cost/cnt);

	delete rt;
	delete c;
	fclose(qfp);
}

void main(int argc, char* argv[])
{
	float s=4;

	if (s==1) 
	{
	//	build_rtree("../ds/lb.txt", "./trees/lb10k32-24", 1024, 32, 24, true);
	//	build_rtree("../ds/lb.txt", "./trees/lb10k16-24", 1024, 16, 24, true);
	//	build_rtree("../ds/lb.txt", "./trees/lb10k8-24", 1024, 8, 24, true);
		build_rtree("../ds/lb.txt", "./trees/lb100k16-24", 1024, 16, 24, true);
	}
	if (s==2) 
	{
		//insert_bitmaps("../ds/lb.txt", "../ds/lb10k.sig32-24", "./trees/lb10k32-24", 1024, 10000);
		//insert_bitmaps("../ds/lb.txt", "../ds/lb10k.sig16-24", "./trees/lb10k16-24", 1024, 10000);
		//insert_bitmaps("../ds/lb.txt", "../ds/lb10k.sig8-24", "./trees/lb10k8-24", 1024, 10000);
		insert_bitmaps("../ds/lb.txt", "../ds/lb100k.sig16-24", "./trees/lb100k16-24", 1024, 10000);
	}
	if (s==2.5) 
	{
		integrity_check("../ds/lb.txt", "../ds/lb10k.sig32-24", "./trees/lb10k32-24", 1024, 10000); 
	}
	if (s==3)
	{
		float qmbr[]={3500, 4500, 3500, 4500};
		float qt[]={1,19};
		Cache *c=new Cache(0, 1024);
		RTree *rt=new RTree("./trees/u1k1k32-32", c);
		float res=rt->agg_qry(qmbr, qt, 20);
		printf("res=%.3f\n", res);
		delete rt;
		delete c;
	}
	if (s==4) 
	{
		//aggq_wrkld("./trees/lb10k32-24", "../qs/lb10k.2.5k-10", 1024, 100+1, 2);
		//aggq_wrkld("./trees/lb10k16-24", "../qs/lb10k.2.5k-10", 1024, 100+1, 2);
		//aggq_wrkld("./trees/lb10k8-24", "../qs/lb10k.2.5k-10", 1024, 100+1, 2);

		//aggq_wrkld("./trees/lb10k16-24", "../qs/lb10k.2.5k-10", 1024, 100+1, 1);
		//aggq_wrkld("./trees/lb10k16-24", "../qs/lb10k.1.5k-10", 1024, 100+1, 1);
		//aggq_wrkld("./trees/lb10k16-24", "../qs/lb10k.0.5k-10", 1024, 100+1, 1);

		//aggq_wrkld("./trees/lb10k8-24", "../qs/lb10k.1.5k-1", 1024, 100+1, 2);
		//aggq_wrkld("./trees/lb10k8-24", "../qs/lb10k.1.5k-20", 1024, 100+1, 2);

		aggq_wrkld("./trees/lb100k16-24", "../qs/lb100k.1.5k-10", 1024, 100+1, 2);
	}
}
